# Program Data
main_directory = 'Videos'
data_directory = '.Gridvid'
user_directory = 'Gridvid'
outputs_directory = 'Program Output'
inputs_directory = 'User Input'

# Image/Video IO
finished_videos_directory = 'Videos'
image_output_directory ='Images'
input_video_directory = 'Videos'
temp_video_directory = '.temporary videos'

# Filenames
video_filename = 'video'

# File Extensions
image_extension = '.png'
